// SPDX-License-Identifier: MIT
pragma solidity 0.8.12;

import "@openzeppelin/contracts/token/ERC20/utils/SafeERC20.sol";
import "@openzeppelin/contracts/utils/Address.sol";

import "./token/Laminor.sol";
import "./TokensVesting.sol";

contract LaminorCrowdsale is Context {
    using SafeMath for uint256;
    using Address for address;
    using SafeERC20 for IERC20;
    using SafeERC20 for Laminor;

    Laminor public LMNToken;
    address public busd = address(0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56);
    address private _owner;

    uint256 public startVesting;
    TokensVesting public vestingContract;

    uint256 public tokenAmountRateOne = 40;
    uint256 public tokenAmountRateTwo = 20;
    uint256 public tokensRaised;
    uint256 private startCrowdsaleTime;
    uint256 public limitPhaseOne; // 100_000_000_000_000_000_000
    uint256 public limitPhaseTwo; // 50_000_000_000_000_000_000
    uint256 public minimumBuyAmount = 10 * 1e18;
    uint256 public BNBPrice = 420;
    bool public isIcoCompleted = false;
    bool public hasIcoPaused = false;

    bool public isIcoFirstRoundCompleted = false;
    uint256 public startSecondRoundTime;

    event LMNTokenBuy(address indexed buyer, uint256 value, uint256 amount);

    modifier whenIcoCompleted() {
        require(isIcoCompleted);
        _;
    }

    modifier onlyOwner() {
        require(owner() == _msgSender(), "Ownable: caller is not the owner");
        _;
    }

    modifier onlyAfterStart() {
        require(
            block.timestamp >= startCrowdsaleTime,
            "Crowdsale: crowdsale not started yet"
        );
        _;
    }

    modifier onlyWhenNotPaused() {
        require(!hasIcoPaused, "Crowdsale: crowdsale has paused");
        _;
    }

    constructor(
        uint256 _startCrowdsaleTime,
        address _LMNToken,
        address payable _vestingContract
    ) {
        require(_startCrowdsaleTime >= 0, "Crowdsale: invalid time provided");
        require(
            _startCrowdsaleTime + 60 >= block.timestamp,
            "Crowdsale: can not start in past"
        );

        require(
            _LMNToken != address(0),
            "Crowdsale: invalid zero address for token provided"
        );

        require(
            _vestingContract != address(0),
            "Crowdsale: invalid zero address for vest provided"
        );
        startCrowdsaleTime = _startCrowdsaleTime;
        vestingContract = TokensVesting(_vestingContract);
        LMNToken = Laminor(_LMNToken);
        _owner = _msgSender();

        limitPhaseOne = 100e9 * (10**LMNToken.decimals());
        limitPhaseTwo = 150e9 * (10**LMNToken.decimals());
    }

    function owner() public view virtual returns (address) {
        return _owner;
    }

    function transferOwnership(address newOwner) public onlyOwner {
        require(newOwner != address(0), "new owner is the zero address");
        _owner = newOwner;
    }
    function setBUSD(address _busd) public onlyOwner {
        require(_busd != address(0));
        busd = _busd;
    }

    function setBNBPrice(uint256 price) public onlyOwner {
        require(price > 0, "Crowdsale: price");
        BNBPrice = price;
    }

    function deposit(uint256 amount) public onlyOwner {
        LMNToken.safeTransferFrom(_msgSender(), address(this), amount);
    }

    function withdrawLMN() public onlyOwner {
        uint256 balance = LMNToken.balanceOf(address(this));
        LMNToken.safeTransfer(_msgSender(), balance);
    }

    function setStartCrowdsaleTime(uint256 _time) public onlyOwner {
        require(_time >= 0, "Crowdsale: invalid time provided");
        require(
            _time + 60 >= block.timestamp,
            "Crowdsale: can not start in past"
        );
        startCrowdsaleTime = _time;
    }

    function timeUntilCrowdsaleStarts() public view returns (uint256) {
        return startCrowdsaleTime;
    }

    function buy(address _token, uint256 _amount)
        public
        onlyAfterStart
        onlyWhenNotPaused
    {
        require(
            tokensRaised <= limitPhaseTwo,
            "Crowdsale: reached the maximum amount"
        );

        require(_token == busd, "Crowdsale: only busd is used for buying");

        uint256 tokensToBuy;

        if (tokensRaised < limitPhaseOne) {
            tokensToBuy = _getTokensAmount(_amount, tokenAmountRateOne);
            require(
                (_amount >= minimumBuyAmount) || (tokensToBuy >= (limitPhaseOne - tokensRaised)),
                "Crowdsale: insufficient balance for buying (minimum 10 busd)"
            );
            if (tokensRaised + tokensToBuy > limitPhaseOne) {
                revert("Crowdsale: exceeds the maximum allowed limit");
            }
        } else if (
            tokensRaised >= limitPhaseOne &&
            tokensRaised < limitPhaseTwo &&
            isIcoFirstRoundCompleted
        ) {
            uint256 secondRoundTime = getStartSecondRound();
            require(secondRoundTime > 0 && block.timestamp >= secondRoundTime, "Crowdsale: second round not started!");
            tokensToBuy = _getTokensAmount(_amount, tokenAmountRateTwo);
            require(
                (_amount >= minimumBuyAmount) || (tokensToBuy >= (limitPhaseTwo - tokensRaised)),
                "Crowdsale: insufficient balance for buying (minimum 10 busd)"
            );
            if (tokensRaised + tokensToBuy > limitPhaseTwo) {
                revert("Crowdsale: exceeds the maximum allowed limit");
            }
        }

        require(
            tokensToBuy > 0,
            "Crowdsale: insufficient balance for buying (minimum 10 busd)"
        );

        IERC20(_token).safeTransferFrom(_msgSender(), owner(), _amount);
        LMNToken.safeTransfer(address(vestingContract), tokensToBuy);

        vestingContract.vest(_msgSender(), tokensToBuy, false);
        emit LMNTokenBuy(_msgSender(), _amount, tokensToBuy);

        tokensRaised += tokensToBuy;
        if (tokensRaised == limitPhaseOne) {
            isIcoFirstRoundCompleted = true;
        }
    }

    function buyNative()
        public
        payable
        onlyAfterStart
        onlyWhenNotPaused
    {
        require(
            tokensRaised <= limitPhaseTwo,
            "Crowdsale: reached the maximum amount"
        );

        uint256 _amount = msg.value;
        uint256 busdAmount = _amount * BNBPrice;

        uint256 tokensToBuy;

        if (tokensRaised < limitPhaseOne) {
            tokensToBuy = _getTokensAmount(busdAmount, tokenAmountRateOne);
            require(
                (busdAmount >= minimumBuyAmount) || (tokensToBuy >= (limitPhaseOne - tokensRaised)),
                "Crowdsale: insufficient balance for buying (minimum 10 busd)"
            );
            if (tokensRaised + tokensToBuy > limitPhaseOne) {
                revert("Crowdsale: exceeds the maximum allowed limit");
            }
        } else if (
            tokensRaised >= limitPhaseOne &&
            tokensRaised < limitPhaseTwo &&
            isIcoFirstRoundCompleted
        ) {
            uint256 secondRoundTime = getStartSecondRound();
            require(secondRoundTime > 0 && block.timestamp >= secondRoundTime, "Crowdsale: second round not started!");
            tokensToBuy = _getTokensAmount(busdAmount, tokenAmountRateTwo);
            require(
                (busdAmount >= minimumBuyAmount) || (tokensToBuy >= (limitPhaseTwo - tokensRaised)),
                "Crowdsale: insufficient balance for buying (minimum 10 busd)"
            );
            if (tokensRaised + tokensToBuy > limitPhaseTwo) {
                revert("Crowdsale: exceeds the maximum allowed limit");
            }
        }

        require(
            tokensToBuy > 0,
            "Crowdsale: insufficient balance for buying (minimum 10 busd)"
        );

        payable(owner()).transfer(_amount);
        LMNToken.safeTransfer(address(vestingContract), tokensToBuy);

        vestingContract.vest(_msgSender(), tokensToBuy, false);
        emit LMNTokenBuy(_msgSender(), _amount, tokensToBuy);

        tokensRaised += tokensToBuy;
        if (tokensRaised == limitPhaseOne) {
            isIcoFirstRoundCompleted = true;
        }
    }

    function _getTokensAmount(uint256 _busdAmount, uint256 _tokenAmountRate)
        internal
        view
        returns (uint256)
    {
        // 1 LMN Token = 1_000_000_000 => 1e9
        // 1 busd       = 1_000_000_000_000_000_000 => 1e18
        // (msg.value*10^18) * (1*10^9) / (1*10^18) => 1 LMN Token

        uint256 oneLMNToken = (10**LMNToken.decimals());
        uint256 amountOfTokens = _busdAmount.mul(oneLMNToken).div(1e18);

        return amountOfTokens.mul(_tokenAmountRate);
    }

    function getTokensAmount(uint256 _busdAmount, uint256 _tokenAmountRate)
        public
        view
        returns (uint256 tokensAmount)
    {
        tokensAmount = _getTokensAmount(_busdAmount, _tokenAmountRate);
    }

    function closeCrowdsale() public onlyOwner {
        isIcoCompleted = true;
        setStartVestingTime();
        vestingContract.startTheVesting(getStartVestingTime());
    }

    function togglePauseCrowdsale() public onlyOwner {
        hasIcoPaused = !hasIcoPaused;
    }

    function getPauseCrowdsaleState() public view returns (bool) {
        return hasIcoPaused;
    }

    function totalCrowdsale() public view returns (uint256) {
        return limitPhaseTwo;
    }

    function startSecondRound(uint256 _time) public onlyOwner {
        require(_time >= 0, "Crowdsale: invalid time provided");
        require(
            _time + 60 >= block.timestamp,
            "Crowdsale: can not start in past"
        );
        startSecondRoundTime = _time;
    }

    function getStartSecondRound() public view returns (uint256) {
        return startSecondRoundTime;
    }

    function withdraw() public whenIcoCompleted onlyOwner {
        (bool success, ) = payable(owner()).call{value: address(this).balance}(
            ""
        );
        require(success);
    }

    function setStartVestingTime() public onlyOwner {
        startVesting = block.timestamp;
    }

    function getStartVestingTime() public view returns (uint256) {
        return startVesting;
    }

    function getTotalBuyedTokens() public view returns (uint256) {
        return tokensRaised;
    }
}