// SPDX-License-Identifier: MIT
pragma solidity 0.8.12;

import "@openzeppelin/contracts/utils/math/SafeMath.sol";
import "@openzeppelin/contracts/token/ERC20/utils/SafeERC20.sol";

import "./token/Laminor.sol";

contract TokensVesting is Ownable {
    using SafeMath for uint256;
    using SafeERC20 for Laminor;

    event NewVest(address indexed _from, address indexed _to, uint256 _value);
    event UnlockVest(address indexed _holder, uint256 _value);
    event RevokeVest(address indexed _holder, uint256 _refund);

    struct Vest {
        uint256 value;
        uint256 releasesCount; // 10%
        uint256 transferred;
        bool revokable;
        bool revoked;
    }

    address public crowdsaleAddress;
    Laminor public LMNToken;
    uint256 public totalVesting;
    uint256 public totalLimit;
    uint256 finishOfVest = 10 * 30 days;
    address[] public vesters;
    mapping(address => Vest) public vests;
    uint256 start;
    uint256 finish;
    uint256 duration = 30 days;

    modifier onlyCrowdsale() {
        require(_msgSender() == crowdsaleAddress);
        _;
    }

    constructor(address _LMNToken) {
        require(
            _LMNToken != address(0),
            "TokenVestings: invalid zero address for token provided"
        );

        LMNToken = Laminor(_LMNToken);
        totalLimit = 150e9 * (10 ** LMNToken.decimals());
    }

    function setCrowdsaleAddress(address _crowdsaleAddress) public onlyOwner {
        require(
            _crowdsaleAddress != address(0),
            "TokenVestings: invalid zero address for crowdsale"
        );
        crowdsaleAddress = _crowdsaleAddress;
    }

    function vest(
        address _to,
        uint256 _value,
        bool _revokable
    ) public onlyCrowdsale {
        require(
            _to != address(0),
            "TokensVesting: invalid zero address for beneficiary!"
        );
        require(_value > 0, "TokensVesting: invalid value for beneficiary!");
        require(totalVesting.add(_value) <= totalLimit, "TokensVesting: total value exeeds total limit!");

        if (vests[_to].value == 0) {
            vests[_to] = Vest({
                value: 0,
                releasesCount: 10,
                transferred: 0,
                revokable: _revokable,
                revoked: false
            });
            vesters.push(_to);
        }

        vests[_to].value += _value;

        totalVesting = totalVesting.add(_value);

        emit NewVest(_msgSender(), _to, _value);
    }

    function revoke(address _holder) public onlyOwner {
        Vest storage vested = vests[_holder];

        require(vested.revokable, "TokensVesting: vested can not get revoked!");
        require(!vested.revoked, "TokensVesting: holder already revoked!");

        uint256 refund = vested.value.sub(vested.transferred);

        totalVesting = totalVesting.sub(refund);
        vested.revoked = true;
        LMNToken.safeTransfer(_msgSender(), refund);

        emit RevokeVest(_holder, refund);
    }

    function beneficiary(address _ben) public view returns (Vest memory) {
        return vests[_ben];
    }

    function startTheVesting(uint256 _start) public onlyCrowdsale {
        require(finish == 0, "TokensVesting: already started!");
        start = _start;
        finish = _start.add(finishOfVest);
    }

    function vestedTokens(address _holder, uint256 _time)
        public
        view
        returns (uint256)
    {
        if (start == 0) {
            return 0;
        }

        Vest memory vested = vests[_holder];
        if (vested.value == 0) {
            return 0;
        }

        return calculateVestedTokens(vested, _time);
    }

    function calculateVestedTokens(Vest memory _vested, uint256 _time)
        private
        view
        returns (uint256)
    {
        if (_time >= finish) {
            return _vested.value;
        }

        uint256 timeLeftAfterStart = block.timestamp.sub(start);
        uint256 availableReleases = timeLeftAfterStart.div(duration);
        uint256 tokensPerRelease = _vested.value.mul(_vested.releasesCount).div(100);

        return availableReleases.mul(tokensPerRelease);
    }

    function unlockVestedTokens() public {
        require(start != 0, "TokensVesting: not started yet");
        Vest storage vested = vests[_msgSender()];
        require(vested.value != 0);
        require(!vested.revoked);

        uint256 vestedAmount = calculateVestedTokens(vested, block.timestamp);
        if (vestedAmount == 0) {
            return;
        }

        uint256 transferable = vestedAmount.sub(vested.transferred);
        if (transferable == 0) {
            return;
        }

        vested.transferred = vested.transferred.add(transferable);
        totalVesting = totalVesting.sub(transferable);
        LMNToken.safeTransfer(_msgSender(), transferable);

        emit UnlockVest(_msgSender(), transferable);
    }
}